library nui_geolocator;

import 'dart:async';
import 'package:flutter/cupertino.dart';
import 'package:flutter/material.dart';
import 'package:geocoding/geocoding.dart';
import 'package:geolocator/geolocator.dart';
import 'package:intl/intl.dart';
import 'package:map_launcher/map_launcher.dart';
//import 'package:location/location.dart';
import 'package:nui_core/nui_core.dart';
import 'package:nui_geolocator/ui/nui_map_selection_dialog.dart';

typedef LocationUpdateCallback = void Function(Position position, AreaCode areaCode);

String formatDistance(double distanceMetre){
  final numberFormat = NumberFormat("#,###.#");
  if (distanceMetre < 1000) {
    if(distanceMetre <= 1){
      return "Right here";
    }
    return numberFormat.format(distanceMetre) + "m";
  }
  return numberFormat.format(distanceMetre/1000) + "km";
}

class NUILatLng extends NUIEnt<NUILatLng>{
  NUILatLng({@required this.latitude, @required this.longitude});
  NUILatLng.empty();
  double latitude;
  double longitude;

  @override
  NUIEntMapper<NUILatLng> mapper() => LatLngEntity();
}

class LatLngEntity extends NUIEntMapper<NUILatLng>{
  @override
  String entityName() => "NUIGeoLatLng";

  @override
  List<NUIEntField> fields() =>[
    NUIEntField(name: "latitude", type: NUIEntType.DOUBLE, setter: (data, value) => data.latitude = value, getter: (data) => data.latitude),
    NUIEntField(name: "longitude", type: NUIEntType.DOUBLE, setter: (data, value) => data.longitude = value, getter: (data) => data.longitude),
  ];

  @override
  NUILatLng newInstance() => NUILatLng.empty();
}

class AreaCode extends NUIEnt<AreaCode> with AreaInfo{
  AreaCode.empty();
  AreaCode({@required this.location, Placemark placemark}){
    addressLine = placemark.street;
    addressLine2 = isNullOrEmpty(placemark.locality) ? null : placemark.subLocality;
    city = isNullOrEmpty(placemark.locality) ? placemark.subLocality : placemark.locality;
    state = placemark.administrativeArea;
    countryCode = placemark.isoCountryCode;
    countryName = placemark.country;
    postalCode = placemark.postalCode;
  }

  NUILatLng location;

  @override
  NUIEntMapper<AreaCode> mapper() => AreaCodeEntity();
}

class AreaCodeEntity extends NUIEntMapper<AreaCode>{
  @override
  String entityName() => "NUIGeoAreaCode";

  @override
  List<NUIEntField> fields(){
    final fields = AreaInfo.fields<AreaCode>();
    fields.add(NUIEntField(name: "location", type: NUIEntType.OBJECT, setter: (data, value) => data.location = asObject(LatLngEntity(), value), getter: (data) => data.location));
    return fields;
  }

  @override
  AreaCode newInstance() => AreaCode.empty();
}

abstract class AreaInfo{
  static List<NUIEntField> fields<T extends AreaInfo>() {
    return [
      NUIEntField<T>(name: "addressLine", type: NUIEntType.STRING, setter: (data, value) => data.addressLine = value, getter: (data) => data.addressLine),
      NUIEntField<T>(name: "addressLine2", type: NUIEntType.STRING, setter: (data, value) => data.addressLine2 = value, getter: (data) => data.addressLine2),
      NUIEntField<T>(name: "city", type: NUIEntType.STRING, setter: (data, value) => data.city = value, getter: (data) => data.city),
      NUIEntField<T>(name: "postalCode", type: NUIEntType.STRING, setter: (data, value) => data.postalCode = value, getter: (data) => data.postalCode),
      NUIEntField<T>(name: "state", type: NUIEntType.STRING, setter: (data, value) => data.state = value, getter: (data) => data.state),
      NUIEntField<T>(name: "countryName", type: NUIEntType.STRING, setter: (data, value) => data.countryName = value, getter: (data) => data.countryName),
      NUIEntField<T>(name: "countryCode", type: NUIEntType.STRING, setter: (data, value) => data.countryCode = value, getter: (data) => data.countryCode),
    ];
  }

  String addressLine;
  String addressLine2;
  String city;
  String postalCode;
  String state;
  String countryName;
  String countryCode;

  String getAreaCodeLineFull(){
    return NUIStringUtil.splitByComma([city, postalCode, state, countryCode], separator: "-");
}

  String getAreaCodeLine(){
    return NUIStringUtil.splitByComma([city, state, countryCode], separator: "-");
  }

  String getAreaDisplay(){
    return NUIStringUtil.splitByComma([city, state, countryName], separator: ", ");
  }

  String getShortLabel(){
    List<String> citySplit = city.split(" ");
    if(citySplit.length > 1){
      if(citySplit.length == 3) {
        return (citySplit[0][0] + "" + citySplit[1][0] + citySplit[2][0]).toUpperCase();
      }
      return (citySplit[0][0] + "" + citySplit[1][0]).toUpperCase();
    }
    else{
      final splited = citySplit[0];
      return (splited[0] + "" +  splited[1] + splited[splited.length - 1]).toUpperCase();
    }
  }
}

class NUIGeoLocator{
  static const MODULE = "NUIGeoLocator";

  NUIGeoLocator._builder(NUIGeoLocatorBuilder builder);
  static NUIGeoLocator _instance;
  static Position _lastKnownPosition;
  static AreaCode _lastKnownAreaCode;

  static const _PREF_LATITUDE = "NUIGeolocator Last Latitude";
  static const _PREF_LONGITUDE = "NUIGeolocator Last Longitude";
  static const _PREF_AREACODE = "NUIGeolocator Last Area Code";

  static NUIGeoLocator get(){
    if(_instance == null) throwNUI(MODULE, "You have not initialized NUIGeoLocator with the builder");
    return _instance;
  }

  Future<Position> getCurrentLocation({LocationAccuracy accuracy, Duration timeOut}) async{
//    Location location = new Location();
//
//    bool _serviceEnabled;
//    PermissionStatus _permissionGranted;
//    LocationData _locationData;
//
//    _serviceEnabled = await location.serviceEnabled();
//    if (!_serviceEnabled) {
//      _serviceEnabled = await location.requestService();
//      if (!_serviceEnabled) {
//        return;
//      }
//    }
//
//    _permissionGranted = await location.hasPermission();
//    if (_permissionGranted == PermissionStatus.denied) {
//      _permissionGranted = await location.requestPermission();
//      if (_permissionGranted != PermissionStatus.granted) {
//        return;
//      }
//    }
//
//    _locationData = await location.getLocation();

    final location = _lastKnownPosition ?? await getCurrentPosition(desiredAccuracy: accuracy ?? LocationAccuracy.high, timeLimit: timeOut ?? Duration(seconds: 10));
    _lastKnownPosition = location;
    if(location != null) {
      _saveLastKnownLatLng(location.latitude, location.longitude);
    }
    return location;
  }

  Future<void> _saveLastKnownLatLng(double lat, double lng) async{
    final saveLat = NUISharedPref.get().save(_PREF_LATITUDE, lat);
    final saveLng = NUISharedPref.get().save(_PREF_LONGITUDE, lng);
    await saveLat;
    await saveLng;
  }

  void launchMapWithCoordinates(BuildContext context, {@required String title, @required NUILatLng latLng}) async{
    final availableMaps = await MapLauncher.installedMaps;

    if(isNullOrEmpty(availableMaps)){
      logNUI("NUIGeolocator", "No maps available in the device");
      return;
    }

    showModalBottomSheet(
      context: context, 
      builder: (context){
        return NUIMapSelectionBottomSheet(
          maps: availableMaps,
          latLng: latLng,
          title: title,
          onMapClick: (map){
            final Coords coordinate = Coords(latLng.latitude, latLng.longitude);
            map.showMarker(
              coords: coordinate,
              title: title,
            );
          },
        );
      }
    );
  }
  
  Future<NUILatLng> getLastKnownLatLng() async{
    if(_lastKnownPosition != null) return NUILatLng(latitude: _lastKnownPosition.latitude, longitude: _lastKnownPosition.longitude);
    final lastLat = await NUISharedPref.get().getDouble(_PREF_LATITUDE, defValue: -1);
    final lastLng = await NUISharedPref.get().getDouble(_PREF_LONGITUDE, defValue: -1);
    logNUI("NUIGeolocator", "Last known lat : $lastLat, lng : $lastLng");
    if(lastLng != null && lastLng != null && lastLat != -1 && lastLng != -1){
      return NUILatLng(latitude: lastLat, longitude: lastLng);
    }
    final position = await getCurrentLocation();
    return NUILatLng(latitude: position.latitude, longitude: position.longitude);
  }

  Future<AreaCode> getLastKnownAreaCode() async{
    if(_lastKnownAreaCode != null) return _lastKnownAreaCode;
    final lastAreacode = await NUISharedPref.get().getObject(_PREF_AREACODE, AreaCodeEntity());
    if(lastAreacode != null){
      return lastAreacode;
    }
    final position = await getLastKnownLatLng();
    return await getAreaCodeFromLocation(position.latitude, position.longitude);
  }

  Future<void> _saveLastKnownAreaCode(AreaCode areaCode) async{
    await NUISharedPref.get().save(_PREF_AREACODE, areaCode);
  }

  Future<List<Placemark>> getPlacemarkFromLocation(double latitude, double longitude) async{
    logNUI(MODULE, "Getting placemarks on (latitude: $latitude, longitude: $longitude)");
    List<Placemark> placemarks = await placemarksFromLocation(lat: latitude, lng: longitude);
    logNUI(MODULE, "${placemarks.length} placemarks obtained");
    if(isNullOrEmpty(placemarks)) return null;
    return placemarks;
  }

  void _printPlacemark(Placemark placemark){
    logNUI(NUIGeoLocator.MODULE, "Placemark, street:${placemark.street}, subLocality:${placemark.subLocality}, locality:${placemark.locality}, administrativeArea:${placemark.administrativeArea}, isoCountryCode:${placemark.isoCountryCode}, countryName:${placemark.country}");
  }

  int _getPlacemarkValue(Placemark placemark){
    int value = 0;
    _printPlacemark(placemark);
    if(!placemark.street.isNullOrEmpty()) value++;
    if(!placemark.subLocality.isNullOrEmpty()) value++;
    if(!placemark.locality.isNullOrEmpty()) value++;
    if(!placemark.administrativeArea.isNullOrEmpty()) value++;
    if(!placemark.isoCountryCode.isNullOrEmpty()) value++;
    if(!placemark.country.isNullOrEmpty()) value++;
    return value;
  }

  Future<AreaCode> getAreaCodeFromLocation(double latitude, double longitude) async{
    final placeMark = await getPlacemarkFromLocation(latitude, longitude);
    if(!placeMark.isNullOrEmpty()){
      Placemark bestChoice;
      int bestValue;
      for(Placemark place in placeMark){
        final curValue = _getPlacemarkValue(place);
        if(bestChoice == null){
          bestChoice = place;
          bestValue = curValue;
        }
        else{
          if(curValue > bestValue){
            bestChoice = place;
            bestValue = curValue;
          }
        }
      }

      logNUI(MODULE, "Best area choice : ");
      _printPlacemark(bestChoice);
      final areaCode = AreaCode(location: NUILatLng(latitude: latitude, longitude: longitude), placemark: bestChoice);
      _lastKnownAreaCode = areaCode;
      _saveLastKnownAreaCode(areaCode);
      return areaCode;
    }
    return null;
  }

  Future<AreaCode> getCurrentLocationInfo({LocationAccuracy accuracy, Duration timeOut}) async{
    logNUI(MODULE, "Getting current position ...");
    Position position = await getCurrentLocation(accuracy: accuracy, timeOut: timeOut ?? Duration(seconds: 2));
    final latlng = position != null ? NUILatLng(latitude: position.latitude, longitude: position.longitude) : await getLastKnownLatLng();
    logNUI(MODULE, "Current position obtained, getting placemarks on (latitude: ${latlng.latitude}, longitude: ${latlng.longitude})");
    final areaCode = await getAreaCodeFromLocation(latlng.latitude, latlng.longitude);
    if(areaCode == null){
      logNUI(MODULE, "No areacode found on (latitude: ${latlng.latitude}, longitude: ${latlng.longitude})");
      return getLastKnownAreaCode();
    }
    return areaCode;
  }

  Future<Position> getLastKnownLocation() async{
    return getLastKnownPosition();
  }

  Future<bool> locationServiceEnabled() async{
    return isLocationServiceEnabled();
  }

  Future<List<Location>> locationsFromAddress(String address) async{
    return locationFromAddress(address);
  }

  Future<List<Placemark>> placemarksFromLocation({@required double lat, @required double lng}) async{
    return placemarkFromCoordinates(lat, lng);
  }

  Future<LocationPermission> checkPermission() async{
    return checkPermission();
  }

  StreamSubscription<Position> listenToLocationUpdate({@required LocationUpdateCallback callback, LocationAccuracy accuracy: LocationAccuracy.high, int distanceTriggerMeters : 5, int interval : 5, Duration timeOut}){
    logNUI(MODULE, "Listening to location update");
    StreamSubscription<Position> positionStream = getPositionStream(
        desiredAccuracy: accuracy ?? LocationAccuracy.high,
        distanceFilter: distanceTriggerMeters ?? 5,
        timeInterval: interval ?? 5,
        timeLimit: timeOut).listen((Position position) async{
          if(position == null){
            logNUI(MODULE, "Location update, position is null when retrieved at ${dateToString(DateTime.now())}");
          }
          else{
            logNUI(MODULE, "Location update, position is (lat:${position.latitude}, lng:${position.longitude}) when retrieved at ${dateToString(DateTime.now())}");
            _lastKnownPosition = position;
            final areaCode = await getAreaCodeFromLocation(position.latitude, position.longitude);
            if(callback != null){
              callback(position, areaCode);
            }
          }
        });
    return positionStream;
  }

  Future<LocationPermission> requestPermission() async{
    return requestPermission();
  }

  Future<double> distanceMetersTo(NUILatLng destination) async{
    final currentLocation = await getLastKnownLatLng();
    return distanceMeters(startLat: currentLocation.latitude, startLng: currentLocation.longitude, endLat: destination.latitude, endLng: destination.longitude);
  }

  double distanceMeters({@required double startLat, @required double startLng, @required double endLat, @required double endLng}){
    return distanceBetween(startLat, startLng, endLat, endLng);
  }

  double bearing({@required double startLat, @required double startLng, @required double endLat, @required double endLng}){
    return bearingBetween(startLat, startLng, endLat, endLng);
  }

  Future<void> checkLocationSetting() async{
    await openAppSettings();
    await openLocationSettings();
  }
}

class NUIGeoLocatorBuilder{
  LocationUpdateCallback _callback;
  bool checkLocation;

  NUIGeoLocatorBuilder setLocationUpdateCallback(LocationUpdateCallback callback){
    this._callback = callback;
    return this;
  }

  NUIGeoLocatorBuilder checkLocationSetting(bool check){
    this.checkLocation = check;
    return this;
  }

  Future<NUIGeoLocator> build() async{
    NUIGeoLocator geo = NUIGeoLocator._builder(this);
    await NUISharedPrefBuilder().build();
    NUIGeoLocator._instance = geo;
    if(checkLocation == true){
      await geo.checkLocationSetting();
    }
    geo.getLastKnownLatLng();
    geo.getLastKnownAreaCode();
    if(_callback != null){
      geo.listenToLocationUpdate(callback: _callback);
    }
    logNUI(NUIGeoLocator.MODULE, "Initialization successful");
    return geo;
  }
}